CC = require 'coffeeclass'

copy =(s,t)->

	for k of s
		if s[k]? and typeof s[k] is 'object'
			if s[k].constructor == Object
				t[k] = {}
				copy s[k], t[k]
			else if s[k].constructor == Array
				t[k] = []
				copy s[k], t[k]
			else if s[k].constructor == Date
				t[k] = new Date s[k]
		else
			t[k] = s[k]
			
module.exports = class Model
	constructor:(params, override)->
		throw new Error "Use #{@constructor.name}.new instead." unless override

		@attrs = {}
		copy params, @attrs
	
		@defaults()
		@validate()
	
		@parse()
		
		#Return the attributes object, NOT the model
		#The model can be accessed through __proto__ (:: for convenience)
		@attrs.__proto__ = @
		@::=@
		return @attrs
	
	validate:->
		@valid = false
		unless @errors = Model.object(@schema)(@attrs)
			@valid = true
	
	defaults:->
		@copy_defaults @schema, @attrs
		
	copy_defaults:(s, t)->
		for k of s
			unless k of t or s[k]?.optional
				t[k] = 	if typeof s[k].default is 'function'
							s[k].default.call(@attrs)
						else
							s[k].default
	#
	
	#
	
	@use:(adapter, options)->
	
		#Treat the object prototype until this call as the model schema
		schema = {}
		for k of @::
			continue if k in ['constructor']
			if @::hasOwnProperty k
				schema[k] = @::[k]
				delete @::[k]
				
		@::schema = schema
		
		options.__proto__ = Config.data[adapter]
		adapter = require('./adapters/' + adapter).call @, options
		
		model = @
		
		@get = CC.wrap {'id'}, (cb)->
			adapter.get @id, cb
			return undefined
		
		@create =(args)->
			fn =(cb)->
				m = model.new @
				m.save(cb)
				return undefined
			fn.__input = args
			return fn
			
		@::save = adapter.save

		@::parse = adapter.parse
		
	@new:(params)->
		new @(params, true).attrs
	
	## Validators
	@with:(schema)->
		try
			return schema.call(@)
		catch e
			if e.type is 'undefined_method' and e.arguments[1] is this
				throw new Error "I don't know how to validate as '#{e.arguments[0]}'."
			else throw e
			
	@_:(options)->
		validate =-> null
		validate.__proto__ = options
		options.type = '_'
		return validate
	
	@array:(cmp...)->
		if typeof cmp[0] is 'object'
			options = cmp.shift()
		else
			options = {}
		
		unless cmp?
			cmp = options
			options = {}
			
		validate = (value) ->
			unless typeof value is 'object' and typeof value.length is 'number'
				return 'must be an array'
			errors = null
			for key of value
				err = []
				for c in cmp
					err.push c(value[key])
					unless err[err.length-1]
						err = null
						break
				if err
					errors ?= {}
					errors[key] = err.join(' or ')
			return errors
		
		options.type = 'array'
		options.cmp = cmp
		validate.__proto__ = options
		validate
	
	@in:(options, cmp)->
		options ?= {}
		unless cmp?
			cmp = options
			options = {}
			
		unless typeof cmp is 'object' and typeof cmp.length is 'number'
			throw new Error 'Comparison for @in must be an array.'
			return
			
		validate = (value) ->
			err = []
			for c in cmp
				if typeof c is 'function'
					err.push c(value)
				else
					err.push "must be `#{c}`" if c != value
					
				unless err[err.length-1]
					err = null
					break
					
			return err.join(' or ') if err
		
		options.type = 'in'
		validate.__proto__ = options
		validate
	
	@object:(options, cmp)->
		options ?= {}
		
		unless cmp?
			cmp = options
			options = {}
			
		unless typeof cmp is 'object'
			throw new Error 'Comparison for @object must be an object.'
			return
		
		if typeof cmp.length is 'number'
			return validators.array options, cmp...
		
			
		regexes = null
		for key of cmp
			if match = /^m\/(.*)\/([gim]*)$/.exec key
				regexes ?= {}
				regexes[key] = new RegExp match[1],match[2]
		
		validate = (value) ->
			unless typeof value is 'object'
				return 'must be an object'
		
			errors = null
			
			for key of value
				continue unless value.hasOwnProperty key
				unless key of cmp
					if cmp.__?
						err = cmp.__(value[key])
						if err
							errors ?= {}
							errors[key] = err
					else if regexes
						err = []
						for rk of regexes
							if regexes[rk].test key
								err.push cmp[rk](value[key])
								unless err[err.length-1]
									err = null
									break
							else
								err.push "not allowed by #{rk}"
						if err
							errors ?= {}
							errors[key] = err.join(' or ')
					else
						errors ?= {}
						errors[key] = 'not allowed'
				else
					unless typeof cmp[key] is 'function'
						throw new Error "Schema error: #{key} is not a validation. Did you forget an @?"
					err = cmp[key](value[key])
					if err
						errors ?= {}
						errors[key] = err
					
			for key of cmp
				continue unless cmp.hasOwnProperty key
				continue if key is '__' or regexes? and key of regexes
				unless key of value or cmp[key].optional?
					errors ?= {}
					errors[key] = 'required'
					
			return errors
		
		options.type = 'object'
		options.cmp = cmp
		validate.__proto__ = options
		validate
		
	@string:(options = {})->
		
		matchers = 
			email: /^[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,4}$/i
			url: /((https?:\/\/)?([-\w]+\.[-\w\.]+)+\w(:\d+)?(\/([-\w\/_\.]*(\?\S+)?)?)*)/
			phone: /^\(?([0-9]{3})\)?[-. ]?([0-9]{3})[-. ]?([0-9]{4})$/
			zip: /^\d{5}(?:-\d{4})?$/
			
		if options.matches?	
			if typeof options.matches is 'string'
				options.match_name ?= options.matches
				options.matches = matchers[options.matches]
			unless typeof options.matches is 'function' and options.matches.constructor is RegExp
				throw new Error "Schema error: @string match `#{options.matches}` is not a RegExp or known matcher."
	
		validate = (value) ->
			unless typeof value is 'string'
				return 'must be a string'
			else if options.matches 
				unless options.matches.test(value)
					return "must match #{options.match_name or options.matches}"
			null
		
		options.coerce ?=(value)->value.toString()
		options.type = 'string'
		validate.__proto__ = options
		validate

	@integer:(options = {})->
		validate = (value) ->
			unless typeof value is 'number' and parseInt(value) == value and not isNaN(value)
				return 'must be an integer'
				
		options.coerce ?=(value)->parseInt(value)
		
		options.type = 'integer'
		validate.__proto__ = options
		validate
		
				
	@boolean: (options = {})->
		
		validate =(value)->
			unless typeof value == 'boolean'
				return "must be a boolean"
				
		options.coerce ?=(value)->value and true
		options.type = 'boolean'
		validate.__proto__ = options
		validate
	
	@date: (options = {})->
		validate = (value)->
			if isNaN (new Date(value).valueOf())
				return "must be a valid date"
					
		options.coerce ?=(value)->new Date(value)
		options.type = 'date'
		validate.__proto__ = options
		validate
		